/* Kernel module to match a string into a packet.
 *
 * Copyright (C) 2000 Emmanuel Roger  <winfield@freegates.be>
 * 
 * ChangeLog
 *	19.02.2002: Gianni Tedesco <gianni@ecsc.co.uk>
 *		Fixed SMP re-entrancy problem using per-cpu data areas
 *		for the skip/shift tables.
 *	02.05.2001: Gianni Tedesco <gianni@ecsc.co.uk>
 *		Fixed kernel panic, due to overrunning boyer moore string
 *		tables. Also slightly tweaked heuristic for deciding what
 * 		search algo to use.
 * 	27.01.2001: Gianni Tedesco <gianni@ecsc.co.uk>
 * 		Implemented Boyer Moore Sublinear search algorithm
 * 		alongside the existing linear search based on memcmp().
 * 		Also a quick check to decide which method to use on a per
 * 		packet basis.
 */

#include <linux/smp.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/file.h>
#include <net/sock.h>
#include <net/tcp.h>
#include <linux/ip.h>

#include <linux/netfilter_ipv4/ip_tables.h>
#include "ipt_http_uri.h"
static int smp_num_cpus=1;

typedef char *(*proc_ipt_search) (char *, char *, int, int);

struct string_per_cpu {
	int *skip;
	int *shift;
	int *len;
};

struct string_per_cpu *bm_string_data=NULL;

/* Boyer Moore Sublinear string search - VERY FAST */
char *search_sublinear (char *needle, char *haystack, int needle_len, int haystack_len) 
{
	int M1, right_end, sk, sh;  
	int ended, j, i;

	int *skip, *shift, *len;
	
	/* use data suitable for this CPU */
	shift=bm_string_data[smp_processor_id()].shift;
	skip=bm_string_data[smp_processor_id()].skip;
	len=bm_string_data[smp_processor_id()].len;
	
	/* Setup skip/shift tables */
	M1 = right_end = needle_len-1;
	for (i = 0; i < BM_MAX_HLEN; i++) skip[i] = needle_len;  
	for (i = 0; needle[i]; i++) skip[ (int) needle[i]] = M1 - i;  // make gcc happy

	for (i = 1; i < needle_len; i++) {   
		for (j = 0; j < needle_len && needle[M1 - j] == needle[M1 - i - j]; j++);  
		len[i] = j;  
	}  

	shift[0] = 1;  
	for (i = 1; i < needle_len; i++) shift[i] = needle_len;  
	for (i = M1; i > 0; i--) shift[len[i]] = i;  
	ended = 0;  
	
	for (i = 0; i < needle_len; i++) {  
		if (len[i] == M1 - i) ended = i;  
		if (ended) shift[i] = ended;  
	}  

	/* Do the search*/  
	while (right_end < haystack_len)
	{
		for (i = 0; i < needle_len && haystack[right_end - i] == needle[M1 - i]; i++);  
		if (i == needle_len) {
			return haystack+(right_end - M1);
		}
		
		sk = skip[ (int) haystack[right_end - i]];   // make gcc happy
		sh = shift[i];
		right_end = max(right_end - i + sk, right_end + sh);  
	}

	return NULL;
}  

/* Linear string search based on memcmp() */
char *search_linear (char *needle, char *haystack, int needle_len, int haystack_len) 
{
	char *k = haystack + (haystack_len-needle_len);
	char *t = haystack;

	// Allen : should not ++ at beginning	
	while ( t <= k ) {
		/* Ron */
#if 0
		if ( memcmp(t, needle, needle_len) == 0 ) return t;
#else
		if ( strnicmp(t, needle, needle_len) == 0 ) return t;
		
#endif
		t++;
	}

	return NULL;
}



int skip_char( char * start, int len, char ch)
{
	int skip =0;


	while( skip < len )
	{
		if( * (start + skip ) == ch)
			skip ++;
		else
			break;
	}

	return skip;
}

int find_char( char * start, int len, char ch)
{
	int skip =0;

	while( skip < len )
	{
		if( * (start + skip ) == ch)
			break;
		else
			skip++;
	}

	return skip;
}


/*

	
	http_start:	should be the start pointer of a http request packet

=================================
	Packet Format				|
=================================

	Method Req-URI Version
	Header: value\r\n
	Header: value\r\n
	Header: value\r\n
	\r\n

	\r=0xd
	\n=0xa
=================================


	return
		1: match
		0: not match
		-1: can't find uri
*/
#define MAX_PACKET 1500
int match_uri( char* http_start, int http_len, char * key, int key_len , u_int16_t invert )
{
	char * req_uri = NULL;
	int req_uri_len=0;
	char * host = NULL;
	int host_len=0;

	char url[MAX_PACKET];
	int url_len;

	char* _key=NULL;
	int _key_len=0;
	
//	char temp_c;
//	int idx;
	int next_ch = 0;
	proc_ipt_search search=search_linear;

	if( http_len < 5) return -1;

/*	// show parameters
	printk("match_uri:\n   key_len = %d\n",key_len);
	printk("   key = %s \n", key);
	printk("   http_len = %d\n",http_len);
*/

/*	// show first 8 characters
	for( idx =0 ; idx < 8 ; idx ++)
		printk("%c" , *(http_start+idx));
	printk("\n\n");
*/
	if(memcmp(http_start, "GET", 3) == 0 )
	{
		//printk("GET\n" );
		next_ch +=3;

	}
	else if(memcmp(http_start, "POST", 4) == 0 )
	{
		//printk("POST\n" );
		next_ch +=4;
	}
	else if(memcmp(http_start, "HEAD", 4) == 0 )
	{
		//printk("HEAD\n" );
		next_ch +=4;
	}
	else
	{
		//printk("match_uri: can't find uri\n");
		return -1;
	}

	// find Request-URI
	next_ch += skip_char( http_start + next_ch , http_len - next_ch , ' '); // skip ths first sp

	if( http_len - next_ch <= 0)
		return -1;

	req_uri_len = find_char( http_start + next_ch, http_len - next_ch, ' ' ); // find the second sp
	req_uri = http_start + next_ch;
	next_ch += req_uri_len;

/*	// show Req-URI
	temp_c= *(req_uri + req_uri_len);
	*(req_uri + req_uri_len) =0;
	printk("Req-uri = %s\n" , req_uri);
	*(req_uri + req_uri_len) =temp_c;
*/	
	// find Host:
	while( next_ch <= http_len)
	{
		next_ch += find_char( http_start + next_ch, http_len - next_ch, '\r' );

		if( 	( (next_ch+1) < http_len) &&  
			(*(http_start + next_ch +1) ) == '\n' )	// skip '\n'
		{
			next_ch+=2;
		}
		else
			break;								// error or partial packet?

		if( 	( (next_ch) < http_len) &&  
			(*(http_start + next_ch ) ) == '\r' )	// reach 2nd '\r\n' --> end of Headers
			break;

		// beginning of a Header
		if( (next_ch+5) > http_len)  // can't be Host header and reach the end of packet
			break;

/*		// show current Header name
		temp_c= *(http_start + next_ch+5 );
		*(http_start +  next_ch+5) = 0;
		printk("-->  %s\n" , http_start +next_ch );
		*(http_start +  next_ch+5) = temp_c;
*/		

		if ( strnicmp( http_start+ next_ch, "Host:", 5) != 0 )	// is Host ?
			continue;		

/*		// show Host
		temp_c= *(http_start +  next_ch+5 );
		*(http_start +  next_ch+5) = 0;
		printk("***  %s\n" , http_start +next_ch );
		*(http_start +  next_ch+5) = temp_c;
*/
		// this is Host header, get value
		next_ch +=5;

		next_ch += skip_char( http_start + next_ch , http_len - next_ch , ' '); // skip the biginning sp
		if( http_len - next_ch <= 0)
			break;

		host_len = find_char( http_start + next_ch, http_len - next_ch, '\r' ); // find the \r\n
		host = http_start + next_ch;

/*		// show Host 
		temp_c= *(host + host_len);
		*(host + host_len) =0;
		printk("Host = %s\n\n" , host);
		*(host + host_len) =temp_c;
*/
		break;
	}


	// merge to url
	url_len = host_len+req_uri_len;
	if( (url_len +1) > MAX_PACKET)	// >_< ... 
	{
		printk("http_uri: (host_len + req_uri_len +1) > MAX_PACKET.   >_<\n" );
		return -1; 
	}

	memcpy( url , host , host_len );
	memcpy( url+host_len , req_uri , req_uri_len );
	url[url_len] = 0;



/*	
	temp_c= *(http_start + http_len -1);
	*(http_start + http_len -1) = 0;
	printk("\n%s\n" , http_start);
	*(http_start + http_len -1) = temp_c;

	printk( "\n\n");
*/


	// skip "http://" of key
	next_ch = find_char( key , key_len , ':'); // skip the biginning sp
	if( 	(next_ch + 3 < key_len) &&
		(*(key + next_ch +1 ) == '/' ) &&
		(*(key + next_ch +2 ) == '/' ) )
		next_ch +=3;
	else
		next_ch =0;
	

	_key = key + next_ch;
	_key_len = key_len - next_ch;

	//printk("URL  = %s\n", url);
	//printk("key2 = %s\n\n", _key);
	//printk("URLL = %d\n", url_len);
	//printk("keyL = %d\n", _key_len);


	/* The sublinear search comes in to its own
	 * on the larger packets */
	if ( (url_len>IPT_STRING_HAYSTACK_THRESH) &&
	  	(_key_len>IPT_STRING_NEEDLE_THRESH) ) {
		if ( url_len < BM_MAX_HLEN ) {
			search=search_sublinear;
			//printk("search_sublinear\n");
			
		}else{
			if (net_ratelimit());
		}
	}

	return ((search(_key, url, _key_len, url_len)!=NULL) ^ invert);

}

static int match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{

	const struct ipt_http_uri_info *info = matchinfo;
	struct iphdr *ip = skb->nh.iph;
	struct tcphdr  *tcp;
	int hlen, nlen;
	char *needle, *haystack;
	char * http_start;
	int ret;
//	proc_ipt_search search=search_linear;

//	printk(" http_uri: match\n\n");


	if ( !ip ) return 0;

	if ( ip->protocol != IPPROTO_TCP)
	{
//		printk(" http_uri: protocol=%d,  not tcp\n", ip->protocol );
		return 0;
	}


	needle=(char *)&info->string;
	haystack=(char *)ip+(ip->ihl*4);

	tcp = (struct tcphdr  *) ((char *)ip+(ip->ihl*4)) ;

	//printk(" http_uri: sport=%d,  dport=%d\n", ntohs(tcp->source ), ntohs( tcp->dest) );

	/* get lenghts, and validate them */
	nlen=info->len;
	hlen=ntohs(ip->tot_len)-(ip->ihl*4) - (tcp->doff *4);
	if ( nlen > hlen ) return 0;

	http_start = (char *)ip+(ip->ihl*4) + (tcp->doff *4) ;

	ret=  match_uri( http_start , hlen , needle , nlen, info->invert);

	if( ret == -1 )
		return 0;

	return ret;

	
#if 0

	/* The sublinear search comes in to its own
	 * on the larger packets */
	if ( (hlen>IPT_STRING_HAYSTACK_THRESH) &&
	  	(nlen>IPT_STRING_NEEDLE_THRESH) ) {
		if ( hlen < BM_MAX_HLEN ) {
			search=search_sublinear;
		}else{
			if (net_ratelimit());
		}
	}
	
    return ((search(needle, haystack, nlen, hlen)!=NULL) ^ info->invert);
#endif
}

static int
checkentry(const char *tablename,
           const struct ipt_ip *ip,
           void *matchinfo,
           unsigned int matchsize,
           unsigned int hook_mask)
{

       if (matchsize != IPT_ALIGN(sizeof(struct ipt_http_uri_info )))
               return 0;

       return 1;
}

void string_freeup_data(void)
{
	int c;
	
	if ( bm_string_data ) 
	{
		for(c=0; c<smp_num_cpus; c++) 
		{
			if ( bm_string_data[c].shift ) kfree(bm_string_data[c].shift);
			if ( bm_string_data[c].skip ) kfree(bm_string_data[c].skip);
			if ( bm_string_data[c].len ) kfree(bm_string_data[c].len);
		}
		kfree(bm_string_data);
	}
}

static struct ipt_match http_uri_match = 
{ 
	{ NULL, NULL }, 
	"http_uri", 
	&match, 
	&checkentry, 
	NULL, 
	THIS_MODULE 
};

static int __init init(void)
{
	int c;
	size_t tlen;
	size_t alen;

	tlen=sizeof(struct string_per_cpu)*smp_num_cpus;
	alen=sizeof(int)*BM_MAX_HLEN;
	
	/* allocate array of structures */
	if ( !(bm_string_data=kmalloc(tlen,GFP_KERNEL)) ) {
		return 0;
	}
	
	memset(bm_string_data, 0, tlen);
	
	/* allocate our skip/shift tables */
	for(c=0; c<smp_num_cpus; c++) {
		if ( !(bm_string_data[c].shift=kmalloc(alen, GFP_KERNEL)) )
			goto alloc_fail;
		if ( !(bm_string_data[c].skip=kmalloc(alen, GFP_KERNEL)) )
			goto alloc_fail;
		if ( !(bm_string_data[c].len=kmalloc(alen, GFP_KERNEL)) )
			goto alloc_fail;
	}
	
	return ipt_register_match(&http_uri_match);

alloc_fail:
	string_freeup_data();
	return 0;
}

static void __exit fini(void)
{
	ipt_unregister_match(&http_uri_match);
	string_freeup_data();
}

module_init(init);
module_exit(fini);


